import argparse
import asyncio
import contextlib as cl
import functools as ft
import hashlib
import io
import os
import os.path as op
import pathlib as pl
import sys

import yaml
import wcpan.logger as wl

from .drive import Drive, DownloadError
from .util import stream_md5sum
from .network import NetworkError


async def verify_upload(drive, local_path, remote_node):
    if local_path.is_dir():
        await verify_upload_directory(drive, local_path, remote_node)
    else:
        await verify_upload_file(drive, local_path, remote_node)


async def verify_upload_directory(drive, local_path, remote_node):
    dir_name = local_path.name

    child_node = drive.get_child_by_name_from_parent(dir_name, remote_node)
    if not child_node:
        wl.ERROR('wcpan.drive.google') << 'not found : {0}'.format(local_path)
        return
    if not child_node.is_folder:
        wl.ERROR('wcpan.drive.google') << 'should be folder : {0}'.format(local_path)
        return

    wl.INFO('wcpan.drive.google') << 'ok : {0}'.format(local_path)

    for child_path in local_path.iterdir():
        await verify_upload(drive, child_path, child_node)


async def verify_upload_file(drive, local_path, remote_node):
    file_name = local_path.name
    remote_path = drive.get_path_by_id(remote_node.id_)
    remote_path = pl.Path(remote_path, file_name)

    child_node = drive.get_child_by_name_from_parent(file_name, remote_node)

    if not child_node:
        wl.ERROR('wcpan.drive.google') << 'not found : {0}'.format(local_path)
        return
    if child_node.is_folder:
        wl.ERROR('wcpan.drive.google') << 'should be file : {0}'.format(local_path)
        return
    if not child_node.available:
        wl.ERROR('wcpan.drive.google') << 'trashed : {0}'.format(local_path)
        return

    with open(local_path, 'rb') as fin:
        local_md5 = stream_md5sum(fin)
    if local_md5 != child_node.md5:
        wl.ERROR('wcpan.drive.google') << 'md5 mismatch : {0}'.format(local_path)
        return

    wl.INFO('wcpan.drive.google') << 'ok : {0}'.format(local_path)


class UploadQueue(object):

    def __init__(self, drive):
        self._drive = drive
        self._lock = asyncio.Semaphore(value=8)
        self._final = asyncio.Condition()
        self._counter = 0
        self._total = 0
        self._failed = []

    async def upload(self, local_path_list, parent_node):
        self._counter = 0
        self._total = sum(self._count_tasks(_) for _ in local_path_list)
        for local_path in local_path_list:
            fn = ft.partial(self._internal_upload, local_path, parent_node)
            self._push(fn)
        await self._wait_for_complete()

    @property
    def failed(self):
        return self._failed

    def _count_tasks(self, local_path):
        total = 1
        for root, folders, files in os.walk(local_path):
            total = total + len(folders) + len(files)
        return total

    async def _wait_for_complete(self):
        async with self._final:
            await self._final.wait()

    async def _internal_upload(self, local_path, parent_node):
        if op.isdir(local_path):
            rv = await self._retry_create_folder(local_path, parent_node)
        else:
            rv = await self._retry_upload_file(local_path, parent_node)
        return rv

    async def _retry_create_folder(self, local_path, parent_node):
        await self._log_begin(local_path)

        folder_name = op.basename(local_path)
        while True:
            try:
                node = await self._drive.create_folder(parent_node, folder_name,
                                                       exist_ok=True)
                break
            except NetworkError as e:
                wl.EXCEPTION('wcpan.drive.google', e)
                if e.status not in ('599',) and e.fatal:
                    self._add_failed(local_path)
                    raise

        await self._log_end(node)

        for child_path in os.listdir(local_path):
            child_path = op.join(local_path, child_path)
            fn = ft.partial(self._internal_upload, child_path, node)
            self._push(fn)

        return node

    async def _retry_upload_file(self, local_path, parent_node):
        await self._log_begin(local_path)

        while True:
            try:
                node = await self._drive.upload_file(local_path, parent_node,
                                                     exist_ok=True)
                break
            except NetworkError as e:
                wl.EXCEPTION('wcpan.drive.google', e)
                if e.status not in ('599',) and e.fatal:
                    self._add_failed(local_path)
                    raise

        await self._log_end(node)

        return node

    def _push(self, runnable):
        loop = asyncio.get_event_loop()
        t = self._do_upload_task(runnable)
        loop.create_task(t)

    async def _do_upload_task(self, runnable):
        async with self._lock:
            await runnable()
            self._counter = self._counter + 1
            if self._counter == self._total:
                async with self._final:
                    self._final.notify()

    def _add_failed(self, local_path):
        self._failed.append(local_path)

    async def _log_begin(self, local_path):
        progress = self._get_progress()
        wl.INFO('wcpan.drive.google') << '{0} begin {1}'.format(progress, local_path)

    async def _log_end(self, remote_node):
        progress = self._get_progress()
        remote_path = await self._drive.get_path(remote_node)
        wl.INFO('wcpan.drive.google') << '{0} end {1}'.format(progress, remote_path)

    def _get_progress(self):
        return '[{0}/{1}]'.format(self._counter, self._total)


class DownloadQueue(object):

    def __init__(self, drive):
        self._drive = drive
        self._lock = asyncio.Semaphore(value=8)
        self._final = asyncio.Condition()
        self._counter = 0
        self._total = 0
        self._failed = []

    async def download(self, node_list, local_path):
        self._counter = 0
        total = (self._count_tasks(_) for _ in node_list)
        total = await asyncio.gather(*total)
        self._total = sum(total)
        for node in node_list:
            if node.is_trashed:
                continue
            fn = ft.partial(self._internal_download, node, local_path)
            self._push(fn)
        await self._wait_for_complete()

    @property
    def failed(self):
        return self._failed

    async def _count_tasks(self, node):
        total = 1
        children = await self._drive.get_children(node)
        count = (self._count_tasks(_) for _ in children)
        count = await asyncio.gather(*count)
        return total + sum(count)

    async def _wait_for_complete(self):
        async with self._final:
            await self._final.wait()

    async def _internal_download(self, node, local_path):
        if node.is_folder:
            rv = await self._create_folder(node, local_path)
        else:
            rv = await self._retry_download_file(node, local_path)
        return rv

    async def _create_folder(self, node, local_path):
        await self._log_begin(node)

        full_path = op.join(local_path, node.name)
        try:
            os.makedirs(full_path, exist_ok=True)
        except Exception as e:
            wl.EXCEPTION('wcpan.drive.google', e)
            self._add_failed(node)

        await self._log_end(full_path)

        children = await self._drive.get_children(node)
        for child in children:
            fn = ft.partial(self._internal_download, child, full_path)
            self._push(fn)

        return full_path

    async def _retry_download_file(self, node, local_path):
        await self._log_begin(node)

        while True:
            try:
                rv = await self._drive.download_file(node, local_path)
                break
            except DownloadError as e:
                wl.EXCEPTION('wcpan.drive.google', e)
                self._add_failed(node)
                raise
            except NetworkError as e:
                wl.EXCEPTION('wcpan.drive.google', e)
                if e.status not in ('599',) and e.fatal:
                    self._add_failed(node)
                    raise

        full_path = op.join(local_path, node.name)
        await self._log_end(full_path)

        return rv

    def _push(self, runnable):
        loop = asyncio.get_event_loop()
        t = self._do_download_task(runnable)
        loop.create_task(t)

    async def _do_download_task(self, runnable):
        async with self._lock:
            await runnable()
            self._counter = self._counter + 1
            if self._counter == self._total:
                async with self._final:
                    self._final.notify()

    def _add_failed(self, local_path):
        self._failed.append(local_path)

    async def _log_begin(self, remote_node):
        progress = self._get_progress()
        remote_path = await self._drive.get_path(remote_node)
        wl.INFO('wcpan.drive.google') << '{0} begin {1}'.format(progress, remote_path)

    async def _log_end(self, local_path):
        progress = self._get_progress()
        wl.INFO('wcpan.drive.google') << '{0} end {1}'.format(progress, local_path)

    def _get_progress(self):
        return '[{0}/{1}]'.format(self._counter, self._total)


async def main(args=None):
    if args is None:
        args = sys.argv

    wl.setup((
        'tornado.access',
        'tornado.application',
        'tornado.general',
        'wcpan.drive.google',
    ))

    args = parse_args(args[1:])
    if not args.action:
        await args.fallback_action()
        return 0

    path = op.expanduser('~/.cache/wcpan/drive/google')

    async with Drive(path) as drive:
        return await args.action(drive, args)


def parse_args(args):
    parser = argparse.ArgumentParser('wdg')

    commands = parser.add_subparsers()

    sync_parser = commands.add_parser('sync', aliases=['s'])
    sync_parser.set_defaults(action=action_sync)

    find_parser = commands.add_parser('find', aliases=['f'])
    add_bool_argument(find_parser, 'id_only')
    add_bool_argument(find_parser, 'include_trash')
    find_parser.add_argument('pattern', type=str)
    find_parser.set_defaults(action=action_find, id_only=False,
                             include_trash=False)

    list_parser = commands.add_parser('list', aliases=['ls'])
    list_parser.set_defaults(action=action_list)
    list_parser.add_argument('id_or_path', type=str)

    tree_parser = commands.add_parser('tree')
    tree_parser.set_defaults(action=action_tree)
    tree_parser.add_argument('id_or_path', type=str)

    dl_parser = commands.add_parser('download', aliases=['dl'])
    dl_parser.set_defaults(action=action_download)
    dl_parser.add_argument('id_or_path', type=str, nargs='+')
    dl_parser.add_argument('destination', type=str)

    ul_parser = commands.add_parser('upload', aliases=['ul'])
    ul_parser.set_defaults(action=action_upload)
    ul_parser.add_argument('source', type=str, nargs='+')
    ul_parser.add_argument('id_or_path', type=str)

    rm_parser = commands.add_parser('remove', aliases=['rm'])
    rm_parser.set_defaults(action=action_remove)
    rm_parser.add_argument('id_or_path', type=str, nargs='+')

    mv_parser = commands.add_parser('rename', aliases=['mv'])
    mv_parser.set_defaults(action=action_rename)
    mv_parser.add_argument('source_path', type=str)
    mv_parser.add_argument('destination_path', type=str)

    sout = io.StringIO()
    parser.print_help(sout)
    fallback = ft.partial(action_help, sout.getvalue())
    parser.set_defaults(action=None, fallback_action=fallback)

    args = parser.parse_args(args)

    return args


def add_bool_argument(parser, name):
    flag = name.replace('_', '-')
    pos_flag = '--' + flag
    neg_flag = '--no-' + flag
    parser.add_argument(pos_flag, dest=name, action='store_true')
    parser.add_argument(neg_flag, dest=name, action='store_false')


async def action_help(message):
    print(message)


async def action_sync(drive, args):
    await drive.sync()
    return 0


async def action_find(drive, args):
    nodes = await drive.find_nodes_by_regex(args.pattern)
    if not args.include_trash:
        nodes = (_ for _ in nodes if _.is_available)
    nodes = (wait_for_value(_.id_, drive.get_path(_)) for _ in nodes)
    nodes = await asyncio.gather(*nodes)
    nodes = dict(nodes)

    if args.id_only:
        for id_ in nodes:
            print(id_)
    else:
        print_as_yaml(nodes)

    return 0


async def action_list(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)
    nodes = await drive.get_children(node)
    nodes = {_.id_: _.name for _ in nodes}
    print_as_yaml(nodes)
    return 0


async def action_tree(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)
    await traverse_node(drive, node, 0)
    return 0


async def action_download(drive, args):
    node_list = (get_node_by_id_or_path(drive, _) for _ in args.id_or_path)
    node_list = await asyncio.gather(*node_list)
    queue_ = DownloadQueue(drive)
    await queue_.download(node_list, args.destination)
    return 0


async def action_upload(drive, args):
    node = await get_node_by_id_or_path(drive, args.id_or_path)
    queue_ = UploadQueue(drive)
    await queue_.upload(args.source, node)
    return 0


async def action_remove(drive, args):
    rv = (trash_node(drive, _) for _ in args.id_or_path)
    rv = await asyncio.gather(*rv)
    rv = filter(None, rv)
    rv = list(rv)
    if rv:
        print_as_yaml(rv)
    return 0


async def action_rename(drive, args):
    node = await drive.rename_node_by_path(args.source_path,
                                           args.destination_path)
    path = await drive.get_path(node)
    return path


async def get_node_by_id_or_path(drive, id_or_path):
    if id_or_path[0] == '/':
        node = await drive.get_node_by_path(id_or_path)
    else:
        node = await drive.get_node_by_id(id_or_path)
    return node


async def traverse_node(drive, node, level):
    if node.is_root:
        print_node('/', level)
    elif level == 0:
        top_path = await drive.get_path(node)
        print_node(top_path, level)
    else:
        print_node(node.name, level)

    if node.is_folder:
        children = await drive.get_children_by_id(node.id_)
        for child in children:
            await traverse_node(drive, child, level + 1)


async def trash_node(drive, id_or_path):
    '''
    :returns: None if succeed, id_or_path if failed
    '''
    node = await get_node_by_id_or_path(drive, id_or_path)
    if not node:
        return id_or_path
    try:
        rv = await drive.trash_node(node)
    except Exception as e:
        return id_or_path
    if not rv:
        return id_or_path
    return None


async def wait_for_value(k, v):
    return k, await v


def print_node(name, level):
    level = ' ' * level
    print(level + name)


def print_as_yaml(data):
    yaml.safe_dump(data, stream=sys.stdout, allow_unicode=True,
                   encoding=sys.stdout.encoding, default_flow_style=False)


main_loop = asyncio.get_event_loop()
exit_code = main_loop.run_until_complete(main())
main_loop.close()
sys.exit(exit_code)



import asyncio
import json
import math
import random
from typing import (Any, AsyncGenerator, AsyncIterator, Callable, Dict,
                    Generator, List, Optional, Text, Tuple, Union)
import urllib.parse as up

import aiohttp
from wcpan.logger import DEBUG, EXCEPTION, INFO, WARNING

from .util import GoogleDriveError, Settings


BACKOFF_FACTOR = 2
BACKOFF_STATUSES = ('403', '500', '502', '503', '504')

ContentProducer = Callable[[], AsyncGenerator[bytes, None]]
ReadableContent = Union[bytes, ContentProducer]


class Network(object):

    def __init__(self, settings: Settings) -> None:
        self._settings = settings
        self._backoff_level = 0
        self._session = None
        self._oauth = None

    async def __aenter__(self) -> 'Network':
        oauth2_info = await self._settings.load_oauth2_info()

        self._session = aiohttp.ClientSession()
        self._oauth = CommandLineGoogleDriveOAuth2(
            self._session,
            oauth2_info['client_id'],
            oauth2_info['client_secret'],
            oauth2_info['redirect_uri'],
            oauth2_info['access_token'],
            oauth2_info['refresh_token'],
        )

        await self._session.__aenter__()
        # FIXME handle exception here, or leak ClientSession
        await self._oauth.__aenter__()
        await self._settings.save_oauth2_info(self._oauth.access_token,
                                              self._oauth.refresh_token)

        return self

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        if self._oauth:
            await self._oauth.__aexit__(exc_type, exc, tb)
            self._oauth = None
        if self._session:
            await self._session.__aexit__(exc_type, exc, tb)
            self._session = None

    async def fetch(self,
            method: Text,
            url: Text,
            args: Dict[Text, Any] = None,
            headers: Dict[Text, Text] = None,
            body: ReadableContent = None,
            raise_internal_error: bool = False,
        ) -> 'Response':
        while True:
            await self._maybe_backoff()
            try:
                rv = await self._do_request(method, url, args, headers, body,
                                            raise_internal_error)
                return rv
            except NetworkError as e:
                if e.status != '599':
                    raise
                if e._response.raise_internal_error:
                    raise
                WARNING('wcpan.drive.google') << str(e)

    async def _do_request(self,
            method: Text,
            url: Text,
            args: Optional[Dict[Text, Any]],
            headers: Optional[Dict[Text, Text]],
            body: Optional[ReadableContent],
            raise_internal_error: bool,
        ) -> 'Response':
        kwargs = {
            'method': method,
            'url': url,
            'headers': self._prepare_headers(headers),
        }
        if args is not None:
            kwargs['params'] = list(normalize_query_string(args))
        if body is not None:
            kwargs['data'] = body if not callable(body) else body()
        if raise_internal_error:
            # do not raise timeout from client
            kwargs['timeout'] = 0

        # retry if access token expired
        while True:
            response = await self._session.request(**kwargs)
            response = Response(response, raise_internal_error)
            if await self._handle_status(response):
                break
            kwargs['headers'] = self._prepare_headers(headers)

        return response

    def _prepare_headers(self,
            headers: Optional[Dict[Text, Text]],
        ) -> Dict[Text, Text]:
        h = {
            'Authorization': 'Bearer {0}'.format(self._oauth.access_token),
        }
        if headers is not None:
            h.update(headers)
        h = {k: v if isinstance(v, (bytes, str)) or v is None else str(v)
             for k, v in h.items()}
        return h

    async def _handle_status(self, response: 'Response') -> 'Response':
        backoff = await backoff_needed(response)
        if backoff:
            self._increase_backoff_level()
        else:
            self._decrease_backoff_level()

        # normal response
        if response.status[0] in ('1', '2', '3'):
            return True

        # need to refresh access token
        if response.status == '401':
            INFO('wcpan.drive.google') << 'refresh token'
            await self._oauth.refresh()
            await self._settings.save_oauth2_info(self._oauth.access_token,
                                                  self._oauth.refresh_token)
            return False

        # otherwise it is an error
        json_ = await response.json()
        raise NetworkError(response, json_, not backoff)

    def _increase_backoff_level(self) -> None:
        self._backoff_level = min(self._backoff_level + 2, 10)

    def _decrease_backoff_level(self) -> None:
        self._backoff_level = max(self._backoff_level - 1, 0)

    async def _maybe_backoff(self) -> None:
        if self._backoff_level <= 0:
            return
        seed = random.random()
        power = 2 ** self._backoff_level
        s_delay = math.floor(seed * power * BACKOFF_FACTOR)
        s_delay = min(100, s_delay)
        DEBUG('wcpan.drive.google') << 'backoff for' << s_delay
        await asyncio.sleep(s_delay)


class Request(object):

    def __init__(self, request: aiohttp.RequestInfo) -> None:
        self._request = request

    @property
    def uri(self) -> Text:
        return self._request.url

    @property
    def method(self) -> Text:
        return self._request.method

    @property
    def headers(self) -> Dict[Text, Text]:
        return self._request.headers


class Response(object):

    def __init__(self,
            response: aiohttp.ClientResponse,
            raise_internal_error: bool,
        ) -> None:
        self._response = response
        self._raise_internal_error = raise_internal_error
        self._request = Request(response.request_info)
        self._status = str(response.status)
        self._parsed_json = False
        self._json = None

    @property
    def status(self) -> Text:
        return self._status

    @property
    def reason(self) -> Text:
        return self._response.reason

    async def json(self) -> Any:
        if self._parsed_json:
            return self._json

        try:
            rv = await self._response.json()
        except aiohttp.ContentTypeError as e:
            EXCEPTION('wcpan.drive.google') << rv
            rv = None
        except ValueError as e:
            EXCEPTION('wcpan.drive.google') << rv
            rv = None

        self._json = rv
        self._parsed_json = True

        return self._json

    def chunks(self) -> AsyncIterator[bytes]:
        return self._response.content.iter_any()

    @property
    def request(self) -> Request:
        return self._request

    @property
    def raise_internal_error(self) -> bool:
        return self._raise_internal_error

    def get_header(self, key: Text) -> Text:
        h = self._response.headers.getall(key)
        return None if not h else h[0]


class NetworkError(GoogleDriveError):

    def __init__(self, response: Response, json_: Any, fatal: bool) -> None:
        self._response = response
        self._message = '{0} {1} - {2}'.format(self.status,
                                               self._response.reason,
                                               json_)
        self._fatal = fatal

    def __str__(self) -> Text:
        return self._message

    @property
    def status(self) -> Text:
        return self._response.status

    @property
    def fatal(self) -> bool:
        return self._fatal

    @property
    def json(self) -> Any:
        return self._json


class CommandLineGoogleDriveOAuth2(object):

    def __init__(self,
            session: aiohttp.ClientSession,
            client_id: Text,
            client_secret: Text,
            redirect_uri: Text,
            access_token: Text = None,
            refresh_token: Text = None
        ) -> None:
        self._session = session
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._access_token = access_token
        self._refresh_token = refresh_token

    async def __aenter__(self) -> 'CommandLineGoogleDriveOAuth2':
        if self._access_token is None:
            await self._fetch_access_token()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        pass

    @property
    def access_token(self) -> Text:
        assert self._access_token is not None
        return self._access_token

    @property
    def refresh_token(self) -> Union[Text, None]:
        return self._refresh_token

    async def refresh(self) -> None:
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        body = up.urlencode({
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'refresh_token': self._refresh_token,
            'grant_type': 'refresh_token',
        })

        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            token = await response.json()
        self._save_token(token)

    def _save_token(self, token: Dict[Text, Any]) -> None:
        self._access_token = token['access_token']
        if 'refresh_token' in token:
            self._refresh_token = token['refresh_token']

    async def _fetch_access_token(self) -> None:
        # get code on success
        code = await self._authorize_redirect()
        token = await self._get_authenticated_user(code=code)
        self._save_token(token)

    async def _authorize_redirect(self) -> Text:
        kwargs = {
            'redirect_uri': self._redirect_uri,
            'client_id': self._client_id,
            'response_type': 'code',
            'scope': ' '.join(self.scopes),
        }

        url = up.urlparse(self.oauth_authorize_url)
        url = up.urlunparse((
            url[0],
            url[1],
            url[2],
            url[3],
            up.urlencode(kwargs),
            url[5],
        ))
        return await self.redirect(url)

    # NOTE Google only
    @property
    def oauth_authorize_url(self) -> Text:
        return 'https://accounts.google.com/o/oauth2/auth'

    # NOTE Google only
    @property
    def oauth_access_token_url(self) -> Text:
        return 'https://accounts.google.com/o/oauth2/token'

    # NOTE Google only
    @property
    def scopes(self) -> List[Text]:
        return [
            'https://www.googleapis.com/auth/drive',
        ]

    # NOTE Google only?
    async def _get_authenticated_user(self, code: Text) -> Dict[Text, Any]:
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = up.urlencode({
            'redirect_uri': self._redirect_uri,
            'code': code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'grant_type': 'authorization_code',
        })
        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            return await response.json()

    # NOTE Use case depends
    async def redirect(self, url: Text) -> Text:
        print(url)
        return input().strip()


async def backoff_needed(response: Response) -> bool:
    if response.status not in BACKOFF_STATUSES:
        return False

    # if it is not a rate limit error, it could be handled immediately
    if response.status == '403':
        msg = await response.json()
        if not msg:
            WARNING('wcpan.drive.google') << '403 with empty error message'
            # probably server problem, backoff for safety
            return True
        domain = msg['error']['errors'][0]['domain']
        if domain != 'usageLimits':
            return False
        INFO('wcpan.drive.google') << msg['error']['message']

    return True


def normalize_query_string(
        qs: Dict[Text, Any],
    ) -> Generator[Tuple[Text, Text], None, None]:
    for key, value in qs.items():
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, (int, float)):
            value = str(value)
        elif not isinstance(value, str):
            raise ValueError('unknown type in query string')
        yield key, value



import functools as ft
import hashlib as hl
import mimetypes
import os
import os.path as op
import re
from typing import (Any, AsyncGenerator, Awaitable, Dict, List, Text, Tuple,
                    Union)

from wcpan.logger import INFO, WARNING, DEBUG
import wcpan.worker as ww

from .api import Client
from .database import Database, Node
from .network import ContentProducer, NetworkError, Response
from .util import Settings, GoogleDriveError, stream_md5sum, FOLDER_MIME_TYPE, CHUNK_SIZE


FILE_FIELDS = 'id,name,mimeType,trashed,parents,createdTime,modifiedTime,md5Checksum,size'
CHANGE_FIELDS = 'nextPageToken,newStartPageToken,changes(fileId,removed,file({0}))'.format(FILE_FIELDS)
EMPTY_MD5SUM = 'd41d8cd98f00b204e9800998ecf8427e'


off_main_thread = ww.off_main_thread_method('_pool')


class Drive(object):

    def __init__(self, settings_path: Text = None) -> None:
        self._settings = Settings(settings_path)
        self._client = None
        self._db = None
        self._pool = None

    async def __aenter__(self) -> 'Drive':
        self._client = Client(self._settings)
        self._db = Database(self._settings)
        self._pool = ww.create_thread_pool()
        await self._client.__aenter__()
        self._db.__enter__()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        self._pool.shutdown()
        self._db.__exit__(exc_type, exc, tb)
        await self._client.__aexit__(exc_type, exc, tb)

    async def sync(self) -> bool:
        INFO('wcpan.drive.google') << 'sync begin'

        try:
            check_point = self._db.get_metadata('check_point')
        except KeyError:
            check_point = '1'

        # first time, get root node
        if check_point == '1':
            rv = await self._client.files.get('root', fields=FILE_FIELDS)
            rv = await rv.json()
            rv['name'] = None
            rv['parents'] = []
            node = Node.from_api(rv)
            self._db.insert_node(node)

        new_start_page_token = None
        changes_list_args = {
            'page_token': check_point,
            'page_size': 1000,
            'restrict_to_my_drive': True,
            'fields': CHANGE_FIELDS,
        }

        while new_start_page_token is None:
            rv = await self._client.changes.list_(**changes_list_args)
            rv = await rv.json()
            next_page_token = rv.get('nextPageToken', None)
            new_start_page_token = rv.get('newStartPageToken', None)
            changes = rv['changes']

            check_point = next_page_token if next_page_token is not None else new_start_page_token

            await self._apply_changes(changes, check_point)
            changes_list_args['page_token'] = check_point

            INFO('wcpan.drive.google') << 'applied' << len(changes) << 'changes'

        INFO('wcpan.drive.google') << 'sync end'
        return True

    @property
    def root_node(self) -> Node:
        return self._db.root_node

    @off_main_thread
    def get_node_by_id(self, node_id: Text) -> Awaitable[Node]:
        return self._db.get_node_by_id(node_id)

    @off_main_thread
    def get_node_by_path(self, path: Text) -> Awaitable[Node]:
        return self._db.get_node_by_path(path)

    @off_main_thread
    def get_path(self, node: Node) -> Awaitable[Text]:
        return self._db.get_path_by_id(node.id_)

    @off_main_thread
    def get_path_by_id(self, node_id: Text) -> Awaitable[Text]:
        return self._db.get_path_by_id(node_id)

    @off_main_thread
    def get_child_by_name_from_parent_id(self,
            name: Text,
            parent_id: Text,
        ) -> Awaitable[Node]:
        return self._db.get_child_by_name_from_parent_id(name, parent_id)

    @off_main_thread
    def get_child_by_name_from_parent(self,
            name: Text,
            parent: Node,
        ) -> Awaitable[Node]:
        return self._db.get_child_by_name_from_parent_id(name, parent.id_)

    @off_main_thread
    def get_children(self, node: Node) -> Awaitable[List[Node]]:
        return self._db.get_children_by_id(node.id_)

    @off_main_thread
    def get_children_by_id(self, node_id: Text) -> Awaitable[List[Node]]:
        return self._db.get_children_by_id(node_id)

    @off_main_thread
    def find_nodes_by_regex(self, pattern: Text) -> Awaitable[List[Node]]:
        return self._db.find_nodes_by_regex(pattern)

    @off_main_thread
    def find_duplicate_nodes(self) -> Awaitable[List[Node]]:
        return self._db.find_duplicate_nodes()

    @off_main_thread
    def find_orphan_nodes(self) -> Awaitable[List[Node]]:
        return self._db.find_orphan_nodes()

    async def download_file_by_id(self, node_id: Text, path: Text) -> bool:
        node = await self.get_node_by_id(node_id)
        return await self.download_file(node, path)

    async def download_file(self, node: Node, path: Text) -> bool:
        # sanity check
        if not node:
            raise ValueError('node is none')
        if node.is_folder:
            raise ValueError('node should be a file')
        if not op.isdir(path):
            raise ValueError('{0} does not exist'.format(path))

        # check if exists
        complete_path = op.join(path, node.name)
        if op.isfile(complete_path):
            return True

        # exists but not a file
        if op.exists(complete_path):
            msg = '{0} exists but is not a file'.format(complete_path)
            raise DownloadError(msg)

        # if the file is empty, no need to download
        if node.size <= 0:
            open(complete_path, 'w').close()
            return True

        # resume download
        tmp_path = complete_path + '.__tmp__'
        if op.isfile(tmp_path):
            offset = op.getsize(tmp_path)
            if offset > node.size:
                msg = ('local file size of `{0}` is greater then remote ({1} > {2})'
                       .format(complete_path, offset, node.size))
                raise DownloadError(msg)
        elif op.exists(tmp_path):
            msg = '{0} exists but is not a file'.format(complete_path)
            raise DownloadError(msg)
        else:
            offset = 0
        range_ = (offset, node.size)

        with open(tmp_path, 'ab') as fout:
            api = self._client.files
            rv = await api.download(file_id=node.id_, range_=range_)
            async for chunk in rv.chunks():
                fout.write(chunk)

        # rename it back if completed
        os.rename(tmp_path, complete_path)

        return True

    async def create_folder(self,
            parent_node: Node,
            folder_name: Text,
            exist_ok: bool = False,
        ) -> Node:
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')

        # do not create again if there is a same file
        node = await self.fetch_node_by_name_from_parent_id(folder_name,
                                                            parent_node.id_)
        if node:
            if exist_ok:
                INFO('wcpan.drive.google') << 'skipped (existing)' << folder_name
                return node
            else:
                raise FileConflictedError(node)

        api = self._client.files
        rv = await api.create_folder(folder_name=folder_name,
                                     parent_id=parent_node.id_)
        rv = await rv.json()
        node = await self.fetch_node_by_id(rv['id'])

        return node

    async def upload_file(self,
            file_path: Text,
            parent_node: Node,
            exist_ok: bool = False,
        ) -> Node:
        # sanity check
        if not parent_node:
            raise UploadError('invalid parent node')
        if not parent_node.is_folder:
            raise UploadError('invalid parent node')
        if not op.isfile(file_path):
            raise UploadError('invalid file path')

        api = self._client.files
        file_name = op.basename(file_path)

        # do not upload if remote exists a same file
        node = await self.fetch_node_by_name_from_parent_id(file_name,
                                                            parent_node.id_)
        if node:
            if exist_ok:
                INFO('wcpan.drive.google') << 'skipped (existing)' << file_path
                return node
            else:
                raise FileConflictedError(node)

        total_file_size = op.getsize(file_path)
        mt, e = mimetypes.guess_type(file_path)
        if total_file_size <= 0:
            rv = await api.create_empty_file(file_name=file_name,
                                             parent_id=parent_node.id_,
                                             mime_type=mt)
            local_md5 = EMPTY_MD5SUM
        else:
            args = {
                'file_path': file_path,
                'file_name': file_name,
                'total_file_size': total_file_size,
                'parent_id': parent_node.id_,
                'mime_type': mt,
            }
            rv, local_md5 = await self._inner_upload_file(**args)

        rv = await rv.json()
        node = await self.fetch_node_by_id(rv['id'])

        if node.md5 != local_md5:
            raise UploadError('md5 mismatch')

        return node

    async def fetch_node_by_name_from_parent_id(self,
            name: Text,
            parent_id: Text,
        ) -> Node:
        safe_name = re.sub(r"[\\']", r"\\\g<0>", name)
        query = "'{0}' in parents and name = '{1}'".format(parent_id,
                                                           safe_name)
        fields = 'files({0})'.format(FILE_FIELDS)
        while True:
            try:
                rv = await self._client.files.list_(q=query, fields=fields)
                break
            except NetworkError as e:
                if e.status == '400':
                    DEBUG('wcpan.drive.google') << 'failed query:' << query
                if e.fatal:
                    raise

        rv = await rv.json()
        files = rv['files']
        if not files:
            return None

        node = Node.from_api(files[0])
        self._db.insert_node(node)
        return node

    async def fetch_node_by_id(self, node_id: Text) -> Node:
        rv = await self._client.files.get(node_id, fields=FILE_FIELDS)
        rv = await rv.json()
        node = Node.from_api(rv)
        self._db.insert_node(node)
        return node

    async def trash_node_by_id(self, node_id: Text) -> Node:
        if node_id == self.root_node.id_:
            return
        await self._client.files.update(node_id, trashed=True)
        node = await self.get_node_by_id(node_id)
        node.is_trashed = True
        self._db.insert_node(node)
        return node

    async def trash_node(self, node: Node) -> Node:
        return await self.trash_node_by_id(node.id_)

    async def rename_node_by_path(self, src_path: Text, dst_path: Text) -> Node:
        src_node = await self.get_node_by_path(src_path)
        if not src_node:
            raise FileNotFoundError(src_path)

        dst_node = await self.get_node_by_path(dst_path)
        # do not support overwriting
        if dst_node and dst_node.is_file:
            raise FileConflictedError(dst_path)

        if dst_node:
            # just move to this folder
            await self._inner_rename_node(src_node, None, dst_node.id_)
        else:
            # move to the parent folder
            dst_folder, dst_name = op.split(dst_path)
            parent_id = await self.get_node_by_path(dst_folder)
            await self._inner_rename_node(src_node, dst_name, parent_id)

        # update local cache
        node = await self.fetch_node_by_id(src_node.id_)
        self._db.insert_node(node)
        return node

    async def _inner_upload_file(self,
            file_path: Text,
            file_name: Text,
            total_file_size: int,
            parent_id: Text,
            mime_type: Text,
        ) -> Tuple[Response, Text]:
        api = self._client.files

        rv = await api.initiate_uploading(file_name=file_name,
                                          total_file_size=total_file_size,
                                          parent_id=parent_id,
                                          mime_type=mime_type)

        url = rv.get_header('Location')

        with open(file_path, 'rb') as fin:
            hasher = hl.md5()
            reader = ft.partial(file_producer, fin, hasher)
            uploader = ft.partial(self._inner_try_upload_file,
                                  url=url, producer=reader,
                                  total_file_size=total_file_size,
                                  mime_type=mime_type)

            retried = False
            offset = 0
            while True:
                ok, rv = await uploader(offset=offset)
                if ok:
                    break
                offset = rv
                fin.seek(offset, os.SEEK_SET)
                retried = True

            if retried:
                fin.seek(0, os.SEEK_SET)
                local_md5 = stream_md5sum(fin)
            else:
                local_md5 = hasher.hexdigest()

        return rv, local_md5

    async def _inner_try_upload_file(self,
            url: Text,
            producer: ContentProducer,
            offset: int,
            total_file_size: int,
            mime_type: Text,
        ) -> Tuple[bool, Union[Response, int]]:
        api = self._client.files

        try:
            rv = await api.upload(url, producer=producer, offset=offset,
                                  total_file_size=total_file_size,
                                  mime_type=mime_type)
            return True, rv
        except NetworkError as e:
            if e.status == '404':
                raise UploadError('the upload session has been expired')
            if e.fatal:
                raise

        while True:
            try:
                rv = await api.get_upload_status(url, total_file_size)
                break
            except NetworkError as e:
                if e.fatal:
                    raise

        if rv.status != '308':
            raise UploadError('invalid upload status')
        rv = rv.get_header('Range')
        if not rv:
            raise UploadError('invalid upload status')
        rv = re.match(r'bytes=(\d+)-(\d+)', rv)
        if not rv:
            raise UploadError('invalid upload status')
        rv = int(rv.group(2))

        return False, rv

    @off_main_thread
    def _apply_changes(self,
            changes: List[Dict[Text, Any]],
            check_point: Text,
        ) -> Awaitable[None]:
        self._db.apply_changes(changes, check_point)

    async def _inner_rename_node(self,
            node: Node,
            name: Text,
            new_parent_id: Text,
        ) -> Response:
        api = self._client.files
        while True:
            try:
                rv = await api.update(file_id=node.id_, name=name,
                                      add_parents=[new_parent_id],
                                      remove_parents=[node.parent_id])
                break
            except NetworkError as e:
                if e.fatal:
                    raise
        return rv


class DownloadError(GoogleDriveError):

    def __init__(self, message: Text) -> None:
        self._message = message

    def __str__(self) -> Text:
        return self._message


class UploadError(GoogleDriveError):

    def __init__(self, message: Text) -> None:
        self._message = message

    def __str__(self) -> Text:
        return self._message


class FileConflictedError(GoogleDriveError):

    def __init__(self, node: Node) -> None:
        self._node = node

    def __str__(self) -> Text:
        return 'remote file already exists: ' + self._node.name


async def file_producer(
        fin: 'file',
        hasher: 'hashlib.hash',
    ) -> AsyncGenerator[bytes, None]:
    while True:
        chunk = fin.read(CHUNK_SIZE)
        if not chunk:
            break
        hasher.update(chunk)
        yield chunk



import datetime as dt
import hashlib
import json
import os
import os.path as op
import re
from typing import Any, BinaryIO, Dict, Text

import yaml


ISO_PATTERN = r'^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(.(\d{3,6}))?(Z|(([+\-])(\d{2}):(\d{2})))$'
CHUNK_SIZE = 64 * 1024
FOLDER_MIME_TYPE = 'application/vnd.google-apps.folder'


class GoogleDriveError(Exception):
    pass


class InvalidDateTimeError(GoogleDriveError):

    def __init__(self, iso: Text) -> None:
        self._iso = iso

    def __str__(self) -> Text:
        return 'invalid ISO date: ' + self._iso


class Settings(object):

    def __init__(self, path: Text) -> None:
        self._path = path
        self._data = {
            'version': 1,
            'save_credentials': False,
            'client_config_backend': 'settings',
            'nodes_database_file': ':memory:',
        }
        self._initialize()

    def _initialize(self) -> None:
        if not self._path:
            return

        # default values for file
        self._data['version'] = 1
        self._data['client_config_backend'] = 'file'
        self._data['client_config_file'] = op.join(self._path, 'client_secret.json')
        self._data['get_refresh_token'] = True
        self._data['save_credentials'] = True
        self._data['save_credentials_backend'] = 'file'
        self._data['save_credentials_file'] = op.join(self._path, 'oauth_token.yaml')
        self._data['nodes_database_file'] = op.join(self._path, 'nodes.db')

        os.makedirs(self._path, exist_ok=True)
        path = op.join(self._path, 'settings.yaml')
        if not op.exists(path):
            rv = yaml.dump(self._data, default_flow_style=False)
            with open(path, 'w') as fout:
                fout.write(rv)
        else:
            with open(path, 'r') as fin:
                rv = yaml.load(fin)
            self._data.update(rv)

    def __getitem__(self, key: Text) -> Any:
        return self._data[key]

    async def load_oauth2_info(self) -> Dict[Text, Any]:
        if 'client_config_file' not in self._data:
            raise ValueError('`client_config_file` not found')

        # load API key
        with open(self._data['client_config_file'], 'r') as fin:
            client = json.load(fin)
        if 'installed' not in client:
            raise ValueError('credential should be an installed application')
        client = client['installed']
        redirect_uri = client['redirect_uris'][0]
        client_id = client['client_id']
        client_secret = client['client_secret']

        # load refresh token
        token_path = self._data['save_credentials_file']
        if not op.isfile(token_path):
            access_token = None
            refresh_token = None
        else:
            with open(token_path, 'r') as fin:
                token = yaml.safe_load(fin)
            if token.get('version', 0) != 1:
                raise ValueError('wrong token file')
            access_token = token['access_token']
            refresh_token = token['refresh_token']

        return {
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': redirect_uri,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }

    async def save_oauth2_info(self,
            access_token: Text,
            refresh_token: Text,
        ) -> None:
        token = {
            'version': 1,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }
        token_path = self._data['save_credentials_file']

        # save refresh token
        rv = yaml.dump(token, default_flow_style=False)
        with open(token_path, 'w') as fout:
            fout.write(rv)


def from_isoformat(iso_datetime: Text) -> dt.datetime:
    rv = re.match(ISO_PATTERN, iso_datetime)
    if not rv:
        raise InvalidDateTimeError(iso_datetime)

    year = int(rv.group(1), 10)
    month = int(rv.group(2), 10)
    day = int(rv.group(3), 10)
    hour = int(rv.group(4), 10)
    minute = int(rv.group(5), 10)
    second = int(rv.group(6), 10)
    if rv.group(8):
        microsecond = rv.group(8).ljust(6, '0')
        microsecond = int(microsecond, 10)
    else:
        microsecond = 0
    tz = rv.group(9)
    if tz == 'Z':
        tz = dt.timezone.utc
    else:
        f = rv.group(11)
        h = int(rv.group(12), 10)
        m = int(rv.group(13), 10)
        tz = dt.timedelta(hours=h, minutes=m)
        if f == '-':
            tz = -tz
        tz = dt.timezone(tz)

    rv = dt.datetime(year, month, day, hour, minute, second, microsecond, tz)
    return rv


def stream_md5sum(input_stream: BinaryIO) -> Text:
    hasher = hashlib.md5()
    while True:
        chunk = input_stream.read(CHUNK_SIZE)
        if not chunk:
            break
        hasher.update(chunk)
    return hasher.hexdigest()



import contextlib as cl
import datetime as dt
import pathlib as pl
import re
import sqlite3
import threading
from typing import Any, Dict, List, Text, Union

from . import util as u


SQL_CREATE_TABLES = [
    '''
    CREATE TABLE metadata (
        key VARCHAR(64) NOT NULL,
        value VARCHAR(4096),
        PRIMARY KEY (key)
    );
    ''',
    '''
    CREATE TABLE nodes (
        id VARCHAR(32) NOT NULL,
        name VARCHAR(4096),
        status VARCHAR(16),
        created DATETIME,
        modified DATETIME,
        PRIMARY KEY (id),
        UNIQUE (id),
        CHECK (status IN ('AVAILABLE', 'TRASH'))
    );
    ''',
    '''
    CREATE TABLE files (
        id VARCHAR(32) NOT NULL,
        md5 VARCHAR(32),
        size BIGINT,
        PRIMARY KEY (id),
        UNIQUE (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',
    '''
    CREATE TABLE parentage (
        parent VARCHAR(32) NOT NULL,
        child VARCHAR(32) NOT NULL,
        PRIMARY KEY (parent, child),
        FOREIGN KEY (child) REFERENCES nodes (id)
    );
    ''',
    'CREATE INDEX ix_parentage_child ON parentage(child);',
    'CREATE INDEX ix_nodes_names ON nodes(name);',
    'PRAGMA user_version = 1;',
]

CURRENT_SCHEMA_VERSION = 1


class DatabaseError(u.GoogleDriveError):

    def __init__(self, message: Text) -> None:
        super(DatabaseError, self).__init__()

        self._message = message

    def __str__(self) -> Text:
        return self._message


class Database(object):

    def __init__(self, settings: u.Settings) -> None:
        self._settings = settings
        self._tls = threading.local()

    def __enter__(self) -> 'Database':
        try:
            self._try_create()
        except sqlite3.OperationalError as e:
            pass

        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('PRAGMA user_version;')
            rv = query.fetchone()
        version = rv[0]

        if CURRENT_SCHEMA_VERSION > version:
            self._migrate(version)

        return self

    def __exit__(self, exc_type, exc, tb) -> bool:
        db = self._get_thread_local_database()
        db.close()

    @property
    def root_id(self) -> Text:
        return self.get_metadata('root_id')

    @property
    def root_node(self) -> 'Node':
        return self.get_node_by_id(self.root_id)

    def get_metadata(self, key: Text) -> Text:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('SELECT value FROM metadata WHERE key = ?;', (key,))
            rv = query.fetchone()
        if not rv:
            raise KeyError
        return rv['value']

    def set_metadata(self, key: Text, value: Text) -> None:
        db = self._get_thread_local_database()
        with ReadWrite(db) as query:
            query.execute('''
                INSERT OR REPLACE INTO metadata
                VALUES (?, ?)
            ;''', (key, value))

    def get_node_by_id(self, node_id: Text) -> Union['Node', None]:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('''
                SELECT id, name, status, created, modified
                FROM nodes
                WHERE id=?
            ;''', (node_id,))
            rv = query.fetchone()
            if not rv:
                return None
            node = dict(rv)

            query.execute('''
                SELECT id, md5, size
                FROM files
                WHERE id=?
            ;''', (node_id,))
            rv = query.fetchone()
            is_folder = rv is None
            node['is_folder'] = is_folder
            node['md5'] = None if is_folder else rv['md5']
            node['size'] = None if is_folder else rv['size']

            query.execute('''
                SELECT parent, child
                FROM parentage
                WHERE child=?
            ;''', (node_id,))
            rv = query.fetchall()
            node['parents'] = None if not rv else [_['parent'] for _ in rv]

            node = Node.from_database(node)
            return node

    def get_node_by_path(self, path: Text) -> Union['Node', None]:
        path = pl.Path(path)
        parts = list(path.parts)
        if parts[0] != '/':
            raise Exception('invalid path')

        node_id = self.root_id
        parts.pop(0)
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            for part in parts:
                query.execute('''
                    SELECT nodes.id AS id
                    FROM parentage
                        INNER JOIN nodes ON parentage.child=nodes.id
                    WHERE parentage.parent=? AND nodes.name=?
                ;''', (node_id, part))
                rv = query.fetchone()
                if not rv:
                    return None
                node_id = rv['id']

            node = self.get_node_by_id(node_id)
        return node

    def get_path_by_id(self, node_id: Text) -> Union[Text, None]:
        parts = []
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            while True:
                query.execute('''
                    SELECT name
                    FROM nodes
                    WHERE id=?
                ;''', (node_id,))
                rv = query.fetchone()
                if not rv:
                    return None

                name = rv['name']
                if not name:
                    parts.insert(0, '/')
                    break
                parts.insert(0, name)

                query.execute('''
                    SELECT parent
                    FROM parentage
                    WHERE child=?
                ;''', (node_id,))
                rv = query.fetchone()
                if not rv:
                    # orphan node
                    break
                node_id = rv['parent']

        path = pl.Path(*parts)
        return str(path)

    def get_child_by_name_from_parent_id(self,
            name: Text,
            parent_id: Text
        ) -> Union['Node', None]:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('''
                SELECT nodes.id AS id
                FROM nodes
                    INNER JOIN parentage ON parentage.child=nodes.id
                WHERE parentage.parent=? AND nodes.name=?
            ;''', (parent_id, name))
            rv = query.fetchone()

        if not rv:
            return None

        node = self.get_node_by_id(rv['id'])
        return node

    def get_children_by_id(self, node_id: Text) -> List['Node']:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('''
                SELECT child
                FROM parentage
                WHERE parent=?
            ;''', (node_id,))
            rv = query.fetchall()

        children = [self.get_node_by_id(_['child']) for _ in rv]
        return children

    def apply_changes(self,
            changes: Dict[Text, Any],
            check_point: Text,
        ) -> None:
        db = self._get_thread_local_database()
        with ReadWrite(db) as query:
            for change in changes:
                is_removed = change['removed']
                if is_removed:
                    inner_delete_node_by_id(query, change['fileId'])
                else:
                    node = Node.from_api(change['file'])
                    inner_insert_node(query, node)

            self.set_metadata('check_point', check_point)

    def insert_node(self, node: 'Node') -> None:
        db = self._get_thread_local_database()
        with ReadWrite(db) as query:
            inner_insert_node(query, node)

        if not node.name:
            self.set_metadata('root_id', node.id_)

    def find_nodes_by_regex(self, pattern: Text) -> List['Node']:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('SELECT id FROM nodes WHERE name REGEXP ?;', (pattern,))
            rv = query.fetchall()
        rv = [self.get_node_by_id(_['id']) for _ in rv]
        return rv

    def find_duplicate_nodes(self) -> List['Node']:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('''
                SELECT nodes.name AS name, parentage.parent as parent_id
                FROM nodes
                    INNER JOIN parentage ON parentage.child=nodes.id
                GROUP BY parentage.parent, nodes.name
                HAVING COUNT(*) > 1
            ;''')
            name_parent_pair = query.fetchall()

            node_id_list = []
            for pair in name_parent_pair:
                query.execute('''
                    SELECT nodes.id AS id
                    FROM nodes
                        INNER JOIN parentage ON parentage.child=nodes.id
                    WHERE parentage.parent=? AND nodes.name=?
                ;''', (pair['parent_id'], pair['name']))
                rv = query.fetchall()
                rv = (_['id'] for _ in rv)
                node_id_list.extend(rv)

            rv = [self.get_node_by_id(_) for _ in node_id_list]
        return rv

    def find_orphan_nodes(self) -> List['Node']:
        db = self._get_thread_local_database()
        with ReadOnly(db) as query:
            query.execute('''
                SELECT nodes.id AS id
                FROM parentage
                    LEFT OUTER JOIN nodes ON parentage.child=nodes.id
                WHERE parentage.parent IS NULL
            ;''')
            rv = query.fetchall()
            rv = [self.get_node_by_id(_['id']) for _ in rv]
        return rv

    def _get_thread_local_database(self) -> sqlite3.Connection:
        db = getattr(self._tls, 'db', None)
        if db is None:
            db = self._open()
            setattr(self._tls, 'db', db)
        return db

    def _open(self) -> sqlite3.Connection:
        path = self._settings['nodes_database_file']
        db = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
        db.row_factory = sqlite3.Row
        db.create_function('REGEXP', 2, sqlite3_regexp)
        return db

    def _try_create(self) -> None:
        db = self._get_thread_local_database()
        with ReadWrite(db) as query:
            for sql in SQL_CREATE_TABLES:
                query.execute(sql)

    def _migrate(self, version) -> None:
        raise NotImplementedError()


class Node(object):

    @staticmethod
    def from_api(data: Dict[Text, Any]) -> 'Node':
        node = Node(data)
        node._initialize_from_api()
        return node

    @staticmethod
    def from_database(data: Dict[Text, Any]) -> 'Node':
        node = Node(data)
        node._initialize_from_database()
        return node

    def __init__(self, data: Dict[Text, Any]) -> None:
        self._data = data

    @property
    def is_root(self) -> bool:
        return self._name is None

    @property
    def is_file(self) -> bool:
        return not self._is_folder

    @property
    def is_folder(self) -> bool:
        return self._is_folder

    @property
    def id_(self) -> Text:
        return self._id

    @property
    def name(self) -> Text:
        return self._name

    @property
    def status(self) -> Text:
        return self._status

    @property
    def is_available(self) -> bool:
        return self._status == 'AVAILABLE'

    @property
    def is_trashed(self) -> bool:
        return self._status == 'TRASH'

    @is_trashed.setter
    def is_trashed(self, trashed: bool) -> None:
        if trashed:
            self._status = 'TRASH'
        else:
            self._status = 'AVAILABLE'

    @property
    def created(self) -> dt.datetime:
        return self._created

    @property
    def modified(self) -> dt.datetime:
        return self._modified

    @property
    def parents(self) -> List[Text]:
        return self._parents

    @property
    def parent_id(self) -> Text:
        if len(self._parents) != 1:
            msg = 'expected only one parent, got: {0}'.format(self._parents)
            raise DatabaseError(msg)
        return self._parents[0]

    @property
    def md5(self) -> Text:
        return self._md5

    @property
    def size(self) -> int:
        return self._size

    def _initialize_from_api(self) -> None:
        data = self._data
        self._id = data['id']
        self._name = data['name']
        self._status = 'TRASH' if data['trashed'] else 'AVAILABLE'
        self._created = u.from_isoformat(data['createdTime'])
        self._modified = u.from_isoformat(data['modifiedTime'])
        self._parents = data.get('parents', None)

        self._is_folder = data['mimeType'] == u.FOLDER_MIME_TYPE
        self._md5 = data.get('md5Checksum', None)
        self._size = data.get('size', None)

    def _initialize_from_database(self) -> None:
        data = self._data
        self._id = data['id']
        self._name = data['name']
        self._status = data['status']
        self._created = data['created']
        self._modified = data['modified']
        self._parents = data.get('parents', None)

        self._is_folder = data['is_folder']
        self._md5 = data['md5']
        self._size = data['size']


class ReadOnly(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, exc_type, exc_value, exc_tb) -> bool:
        self._cursor.close()


class ReadWrite(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, exc_type, exc_value, exc_tb) -> bool:
        if exc_type is None:
            self._db.commit()
        else:
            self._db.rollback()
        self._cursor.close()


def inner_insert_node(query: sqlite3.Cursor, node: Node) -> None:
    # add this node
    query.execute('''
        INSERT OR REPLACE INTO nodes
        (id, name, status, created, modified)
        VALUES
        (?, ?, ?, ?, ?)
    ;''', (node.id_, node.name, node.status, node.created, node.modified))

    # add file information
    if not node.is_folder:
        query.execute('''
            INSERT OR REPLACE INTO files
            (id, md5, size)
            VALUES
            (?, ?, ?)
        ;''', (node.id_, node.md5, node.size))

    # add parentage
    if node.parents:
        query.execute('''
            DELETE FROM parentage
            WHERE child=?
        ;''', (node.id_,))

        for parent in node.parents:
            query.execute('''
                INSERT OR REPLACE INTO parentage
                (parent, child)
                VALUES
                (?, ?)
            ;''', (parent, node.id_))


def inner_delete_node_by_id(query: sqlite3.Cursor, node_id: Text) -> None:
    # disconnect parents
    query.execute('''
        DELETE FROM parentage
        WHERE child=? OR parent=?
    ;''', (node_id, node_id))

    # remove from files
    query.execute('''
        DELETE FROM files
        WHERE id=?
    ;''', (node_id,))

    # remove from nodes
    query.execute('''
        DELETE FROM nodes
        WHERE id=?
    ;''', (node_id,))


def sqlite3_regexp(pattern: Text, cell: Union[Text, None]) -> bool:
    if cell is None:
        # root node
        return False
    return re.search(pattern, cell, re.I) is not None


def initialize() -> None:
    sqlite3.register_adapter(dt.datetime, lambda _: _.isoformat())
    sqlite3.register_converter('DATETIME', to_dt)


def to_dt(raw_datetime):
    s = raw_datetime.decode('utf-8')
    return u.from_isoformat(s)


initialize()



from setuptools import setup


setup(
        name='wcpan.drive.google',
        version='2.0.1',
        author='Wei-Cheng Pan',
        author_email='legnaleurc@gmail.com',
        description='Asynchronous Google Drive API.',
        url='https://github.com/legnaleurc/wcpan.drive.google',
        packages=[
            'wcpan.drive.google',
        ],
        install_requires=[
            'PyYAML',
            'aiohttp >= 3',
            'wcpan.logger',
            'wcpan.worker >= 2',
        ],
        classifiers=[
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.6',
        ])
